# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
"""Disabled as of 2.1.6 but it does seem to work.Except with CD
"""
# TODO Fix dearpygui module not found
# import dearpygui.dearpygui as dpg
#
# import sys
# this = sys.modules[__name__]
#
# for i in dir(dpg):
#     setattr(this, i, getattr(dpg, i))
